@@ -6,7 +6,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2016	PgPool Global Development Group
+ * Copyright (c) 2003-2015	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -279,23 +279,7 @@ int pool_virtual_master_db_node_id(void)
 
 	if (sc->query_context)
 	{
-		int node_id = sc->query_context->virtual_master_node_id;
-
-		if (STREAM)
-		{
-			 /*
-			  * Make sure that virtual_master_node_id is either primary node
-			  * id or load balance node id.  If not, it is likely that
-			  * virtual_master_node_id is not set up yet. Let's use the
-			  * primary node id.
-			  */
-			if (node_id != sc->load_balance_node_id && node_id != PRIMARY_NODE_ID)
-			{
-				node_id = PRIMARY_NODE_ID;
-			}
-		}
-
-		return node_id;
+		return sc->query_context->virtual_master_node_id;
 	}
 
 	/*
@@ -495,7 +479,7 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 			}
 		}
 	}
-	else if (REPLICATION || PARALLEL_MODE)
+	else if (REPLICATION)
 	{
 		if (pool_config->load_balance_mode &&
 			is_select_query(node, query) &&
@@ -685,7 +669,7 @@ POOL_STATUS pool_send_and_wait(POOL_QUERY_CONTEXT *query_context,
 		}
 
 		per_node_statement_log(backend, i, string);
-
+		stat_count_up(i, query_context->parse_tree);
 		send_simplequery_message(CONNECTION(backend, i), len, string, MAJOR(backend));
 	}
 
@@ -746,7 +730,7 @@ POOL_STATUS pool_send_and_wait(POOL_QUERY_CONTEXT *query_context,
  */
 POOL_STATUS pool_extended_send_and_wait(POOL_QUERY_CONTEXT *query_context,
 										char *kind, int len, char *contents,
-										int send_type, int node_id)
+										int send_type, int node_id, bool nowait)
 {
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_CONNECTION *frontend;
@@ -864,6 +848,17 @@ POOL_STATUS pool_extended_send_and_wait(POOL_QUERY_CONTEXT *query_context,
 			per_node_statement_log(backend, i, msgbuf);
 		}
 
+		/* Set sync map so that we could wait for response from appropreate node */
+		pool_set_sync_map(i);
+		ereport(DEBUG1,
+				(errmsg("pool_send_and_wait: pool_set_sync_map: %d", i)));
+
+		/* if Execute message, count up stats count */
+		if (*kind == 'E')
+		{
+			stat_count_up(i, query_context->parse_tree);
+		}
+
 		send_extended_protocol_message(backend, i, kind, str_len, str);
 	}
 
@@ -875,47 +870,50 @@ POOL_STATUS pool_extended_send_and_wait(POOL_QUERY_CONTEXT *query_context,
 			str = query_context->original_query;
 	}
 
-	/* Wait for response */
-	for (i=0;i<NUM_BACKENDS;i++)
+	if (!nowait)
 	{
-		if (!VALID_BACKEND(i))
-			continue;
-		else if (send_type < 0 && i == node_id)
-			continue;
-		else if (send_type > 0 && i != node_id)
-			continue;
-
-		/*
-		 * If in master/slave mode, we do not send COMMIT/ABORT to
-		 * slaves/standbys if it's in I(idle) state.
-		 */
-		if (is_commit && MASTER_SLAVE && !IS_MASTER_NODE_ID(i) && TSTATE(backend, i) == 'I')
+		/* Wait for response */
+		for (i=0;i<NUM_BACKENDS;i++)
 		{
-			continue;
+			if (!VALID_BACKEND(i))
+				continue;
+			else if (send_type < 0 && i == node_id)
+				continue;
+			else if (send_type > 0 && i != node_id)
+				continue;
+
+			/*
+			 * If in master/slave mode, we do not send COMMIT/ABORT to
+			 * slaves/standbys if it's in I(idle) state.
+			 */
+			if (is_commit && MASTER_SLAVE && !IS_MASTER_NODE_ID(i) && TSTATE(backend, i) == 'I')
+			{
+				continue;
+			}
+
+			if (is_begin_read_write)
+			{
+				if (REAL_PRIMARY_NODE_ID == i)
+					str = query_context->original_query;
+				else
+					str = query_context->rewritten_query;
+			}
+
+			wait_for_query_response_with_trans_cleanup(frontend,
+													   CONNECTION(backend, i),
+													   MAJOR(backend),
+													   MASTER_CONNECTION(backend)->pid,
+													   MASTER_CONNECTION(backend)->key);
+
+			/*
+			 * Check if some error detected.  If so, emit
+			 * log. This is useful when invalid encoding error
+			 * occurs. In this case, PostgreSQL does not report
+			 * what statement caused that error and make users
+			 * confused.
+			 */		
+			per_node_error_log(backend, i, str, "pool_send_and_wait: Error or notice message from backend: ", true);
 		}
-
-		if (is_begin_read_write)
-		{
-			if (REAL_PRIMARY_NODE_ID == i)
-				str = query_context->original_query;
-			else
-				str = query_context->rewritten_query;
-		}
-
-        wait_for_query_response_with_trans_cleanup(frontend,
-                                                   CONNECTION(backend, i),
-                                                   MAJOR(backend),
-                                                   MASTER_CONNECTION(backend)->pid,
-                                                   MASTER_CONNECTION(backend)->key);
-
-		/*
-		 * Check if some error detected.  If so, emit
-		 * log. This is useful when invalid encoding error
-		 * occurs. In this case, PostgreSQL does not report
-		 * what statement caused that error and make users
-		 * confused.
-		 */		
-		per_node_error_log(backend, i, str, "pool_send_and_wait: Error or notice message from backend: ", true);
 	}
 
 	if(rewritten_begin)
@@ -945,7 +943,7 @@ static POOL_DEST send_to_where(Node *node, char *query)
 #define AccessExclusiveLock		8		/* ALTER TABLE, DROP TABLE, VACUUM
 										 * FULL, and unqualified LOCK TABLE */
 
-/* From 9.0 include/nodes/node.h */
+/* From 9.5 include/nodes/node.h ("TAGS FOR STATEMENT NODES" part) */
 	static NodeTag nodemap[] = {
 		T_PlannedStmt,
 		T_InsertStmt,
@@ -958,9 +956,7 @@ static POOL_DEST send_to_where(Node *node, char *query)
 		T_SetOperationStmt,
 		T_GrantStmt,
 		T_GrantRoleStmt,
-		/*
-		T_AlterDefaultPrivilegesStmt,	Our parser does not support yet
-		*/
+		T_AlterDefaultPrivilegesStmt,
 		T_ClosePortalStmt,
 		T_ClusterStmt,
 		T_CopyStmt,
@@ -973,9 +969,7 @@ static POOL_DEST send_to_where(Node *node, char *query)
 		T_IndexStmt,	/* CREATE INDEX */
 		T_CreateFunctionStmt,
 		T_AlterFunctionStmt,
-		/*
-		T_DoStmt,		Our parser does not support yet
-		*/
+		T_DoStmt,
 		T_RenameStmt,	/* ALTER AGGREGATE etc. */
 		T_RuleStmt,		/* CREATE RULE */
 		T_NotifyStmt,
@@ -989,6 +983,7 @@ static POOL_DEST send_to_where(Node *node, char *query)
 		T_DropdbStmt,
 		T_VacuumStmt,
 		T_ExplainStmt,
+		T_CreateTableAsStmt,
 		T_CreateSeqStmt,
 		T_AlterSeqStmt,
 		T_VariableSetStmt,		/* SET */
@@ -1024,6 +1019,8 @@ static POOL_DEST send_to_where(Node *node, char *query)
 		T_ReassignOwnedStmt,
 		T_CompositeTypeStmt,	/* CREATE TYPE */
 		T_CreateEnumStmt,
+		T_CreateRangeStmt,
+		T_AlterEnumStmt,
 		T_AlterTSDictionaryStmt,
 		T_AlterTSConfigurationStmt,
 		T_CreateFdwStmt,
@@ -1033,9 +1030,22 @@ static POOL_DEST send_to_where(Node *node, char *query)
 		T_CreateUserMappingStmt,
 		T_AlterUserMappingStmt,
 		T_DropUserMappingStmt,
-		/*
-		T_AlterTableSpaceOptionsStmt,	Our parser does not support yet
-		*/
+		T_AlterTableSpaceOptionsStmt,
+		T_AlterTableMoveAllStmt,
+		T_SecLabelStmt,
+		T_CreateForeignTableStmt,
+		T_ImportForeignSchemaStmt,
+		T_CreateExtensionStmt,
+		T_AlterExtensionStmt,
+		T_AlterExtensionContentsStmt,
+		T_CreateEventTrigStmt,
+		T_AlterEventTrigStmt,
+		T_RefreshMatViewStmt,
+		T_ReplicaIdentityStmt,
+		T_AlterSystemStmt,
+		T_CreatePolicyStmt,
+		T_AlterPolicyStmt,
+		T_CreateTransformStmt,
 	};
 
 	if (bsearch(&nodeTag(node), nodemap, sizeof(nodemap)/sizeof(nodemap[0]),
@@ -1318,7 +1328,8 @@ char *pool_get_query_string(void)
 }
 
 /*
- * Return true if the query is:
+ * Returns true if the query is one of:
+ *
  * SET TRANSACTION ISOLATION LEVEL SERIALIZABLE or
  * SET SESSION CHARACTERISTICS AS TRANSACTION ISOLATION LEVEL SERIALIZABLE or
  * SET transaction_isolation TO 'serializable'
@@ -1332,8 +1343,7 @@ bool is_set_transaction_serializable(Node *node)
 		return false;
 
 	if (((VariableSetStmt *)node)->kind == VAR_SET_VALUE &&
-		(!strcmp(((VariableSetStmt *)node)->name, "transaction_isolation") ||
-                 !strcmp(((VariableSetStmt *)node)->name, "default_transaction_isolation")))
+		!strcmp(((VariableSetStmt *)node)->name, "transaction_isolation"))
 	{
 		List *options = ((VariableSetStmt *)node)->args;
 		foreach(list_item, options)
