@@ -6,7 +6,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2016	PgPool Global Development Group
+ * Copyright (c) 2003-2015	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -35,6 +35,9 @@ static POOL_SESSION_CONTEXT session_context_d;
 static POOL_SESSION_CONTEXT *session_context = NULL;
 static void GetTranIsolationErrorCb(void *arg);
 static void init_sent_message_list(void);
+static POOL_PENDING_MESSAGE *copy_pending_message(POOL_PENDING_MESSAGE *messag);
+static void dump_sent_message(char *caller, POOL_SENT_MESSAGE *m);
+static char *dump_sync_map(void);
 
 /*
  * Initialize per session context
@@ -42,9 +45,6 @@ static void init_sent_message_list(void);
 void pool_init_session_context(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend)
 {
 	session_context = &session_context_d;
-	ProcessInfo *process_info;
-	int node_id;
-	int i;
 
 	/* Get Process context */
 	session_context->process_context = pool_get_process_context();
@@ -73,33 +73,22 @@ void pool_init_session_context(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *
 	/* Initialize sent message list */
 	init_sent_message_list();
 
-	process_info = pool_get_my_process_info();
-
-	if (!process_info)
-		ereport(ERROR,
-				(errmsg("failed to get process info for current process")));
-
 	/* Choose load balancing node if necessary */
 	if (pool_config->load_balance_mode)
 	{
-		node_id = select_load_balancing_node();
-	}
-	else
-	{
-		node_id = STREAM? PRIMARY_NODE_ID: MASTER_NODE_ID;
-	}
+		ProcessInfo *process_info = pool_get_my_process_info();
+		if (!process_info)
+			ereport(ERROR,
+					(errmsg("failed to get process info for current process")));
 
-	session_context->load_balance_node_id = node_id;
+		session_context->load_balance_node_id = 
+			process_info->connection_info->load_balancing_node =
+			select_load_balancing_node();
 
-	for (i=0;i<NUM_BACKENDS;i++)
-	{
-		pool_coninfo(session_context->process_context->proc_id,
-					 pool_pool_index(), i)->load_balancing_node = node_id;
-	}
-
-	ereport(DEBUG1,
+		ereport(DEBUG1,
 			(errmsg("initializing session context"),
-			 errdetail("selected load balancing node: %d", node_id)));
+				 errdetail("selected load balancing node: %d", backend->info->load_balancing_node)));
+	}
 
 	/* Unset query is in progress */
 	pool_unset_query_in_progress();
@@ -137,6 +126,15 @@ void pool_init_session_context(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *
 		session_context->query_cache_array = pool_create_query_cache_array();
 		session_context->num_selects = 0;
 	}
+
+	/* Clear sync map */
+	pool_clear_sync_map();
+
+	/* Unset pending response */
+	pool_unset_pending_response();
+
+	/* Initialize pending message list */
+	pool_pending_messages_init();
 }
 
 /*
@@ -369,6 +367,8 @@ void pool_sent_message_destroy(POOL_SENT_MESSAGE *message)
 	bool in_progress;
 	POOL_QUERY_CONTEXT *qc = NULL;
 
+	dump_sent_message("pool_sent_message_destroy", message);
+
 	in_progress = pool_is_query_in_progress();
 
 	if (message)
@@ -422,6 +422,12 @@ void pool_clear_sent_message_list(void)
 	}
 }
 
+static void dump_sent_message(char *caller, POOL_SENT_MESSAGE *m)
+{
+	ereport(DEBUG1,
+			(errmsg("called by %s: sent message: address: %x kind: %c name: =%s=", caller, m, m->kind, m->name)));
+}
+
 /*
  * Create a sent message
  * kind: one of 'P':Parse, 'B':Bind or'Q':Query(PREPARE)
@@ -463,6 +469,8 @@ void pool_add_sent_message(POOL_SENT_MESSAGE *message)
 	POOL_SENT_MESSAGE *old_msg;
 	POOL_SENT_MESSAGE_LIST *msglist;
 
+	dump_sent_message("pool_add_sent_message", message);
+
 	if (!message)
 	{
 		ereport(DEBUG1,
@@ -472,6 +480,20 @@ void pool_add_sent_message(POOL_SENT_MESSAGE *message)
 	}
 
 	old_msg = pool_get_sent_message(message->kind, message->name);
+
+	if (old_msg == message)
+	{
+		/*
+		 * It is likely caller tries to add the exact same message previously
+		 * added. We should ignore this because pool_remove_sent_message()
+		 * will free memory allocated in the message.
+		 */
+		ereport(DEBUG1,
+				(errmsg("adding sent message to list"),
+				 errdetail("adding exactly same message is prohibited")));
+		return;
+	}
+
 	msglist = &session_context->message_list;
 
 	if (old_msg)
@@ -848,48 +870,325 @@ bool can_query_context_destroy(POOL_QUERY_CONTEXT *qc)
 }
 
 /*
- * Set protocol major version number
+ * Set sync map
  */
-void pool_set_major_version(int major)
+void pool_set_sync_map(int node_id)
 {
-	if (session_context)
-	{
-		session_context->major = major; 
-	}
+	if (!session_context)
+		ereport(ERROR,
+				(errmsg("pool_set_sync_map: session context is not initialized")));
+
+	session_context->sync_map[node_id] = true;
 }
 
 /*
- * Get protocol major version number
+ * Check if sync map is set
  */
-int pool_get_major_version(void)
+bool pool_is_set_sync_map(int node_id)
 {
-	if (session_context)
-	{
-		return session_context->major;
-	}
-	return PROTO_MAJOR_V3;
+	if (!session_context)
+		ereport(ERROR,
+				(errmsg("pool_is_set_sync_map: session context is not initialized")));
+
+	return session_context->sync_map[node_id];
 }
 
 /*
- * Set protocol minor version number
+ * Get nth valid node id from sync map
  */
-void pool_set_minor_version(int minor)
+int pool_get_nth_sync_map(int nth)
 {
-	if (session_context)
+	int i;
+	int cnt = 0;
+
+	for (i = 0; i < NUM_BACKENDS; i++)
 	{
-		session_context->minor = minor; 
+		if (!VALID_BACKEND(i))
+			continue;
+
+		if (session_context->sync_map[i])
+		{
+			if (cnt == nth)
+				return i;	/* found nth valid node id */
+		}
+
+		cnt++;
 	}
+
+	return -1;		/* no valid node id found */
 }
 
 /*
- * Get protocol minor version number
+ * Clear sync map
  */
-int pool_get_minor_version(void)
+void pool_clear_sync_map(void)
 {
-	if (session_context)
-	{
-		return session_context->minor;
-	}
-	return 0;
+	if (!session_context)
+		ereport(ERROR,
+				(errmsg("pool_clear_sync_map: session context is not initialized")));
+
+	memset(&session_context->sync_map, 0, sizeof(session_context->sync_map));
 }
 
+/*
+ * Check whether we should 1) ignore sync map, 2) consult sync map, or 3) we
+ * cannot use sync map because it's empty.
+ */
+bool pool_use_sync_map(void)
+{
+	int i;
+
+	if (!session_context)
+		return POOL_IGNORE_SYNC_MAP;
+
+	if (STREAM && !pool_is_query_in_progress() && pool_is_doing_extended_query_message())
+	{
+		for (i=0;i<NUM_BACKENDS;i++)
+		{
+			if (pool_is_set_sync_map(i))
+			{
+				ereport(DEBUG1,
+						(errmsg("pool_use_sync_map: we can use sync map: %s", dump_sync_map())));
+				return POOL_SYNC_MAP_IS_VALID;	/* yes, we can use sync map */
+			}
+		}
+		ereport(DEBUG1,
+				(errmsg("pool_use_sync_map: we cannot use sync map because all map entries are false")));
+		return POOL_SYNC_MAP_EMPTY;	/* no, we cannot use sync map */
+	}
+
+	ereport(DEBUG1,
+			(errmsg("pool_use_sync_map: we cannot use sync map because STREAM: %d query in progress: %d doing extended query: %d", STREAM, pool_is_query_in_progress(), pool_is_doing_extended_query_message())));
+
+	return POOL_IGNORE_SYNC_MAP;
+}
+
+static char *dump_sync_map(void)
+{
+	static char mapstr[MAX_NUM_BACKENDS+1];
+	int i;
+
+	memset(mapstr, 0, sizeof(mapstr));
+
+	for (i=0;i<NUM_BACKENDS;i++)
+	{
+		mapstr[i] = session_context->sync_map[i]?'1':'0';
+	}
+	return mapstr;
+}
+
+/*
+ * Set pending response
+ */
+void pool_set_pending_response(void)
+{
+	if (!session_context)
+		ereport(ERROR,
+				(errmsg("pool_set_pending_response: session context is not initialized")));
+
+	session_context->is_pending_response = true;
+}
+
+/*
+ * Unset pending response
+ */
+void pool_unset_pending_response(void)
+{
+	if (!session_context)
+		ereport(ERROR,
+				(errmsg("pool_unset_pending_response: session context is not initialized")));
+
+	session_context->is_pending_response = false;
+}
+
+/*
+ * Returns true if pending response is set
+ */
+bool pool_is_pending_response(void)
+{
+	if (!STREAM)
+		return false;
+
+	if (!session_context)
+		ereport(ERROR,
+				(errmsg("pool_is_pending_response: session context is not initialized")));
+
+	return session_context->is_pending_response;
+}
+
+
+/*
+ * Initialize pending message list
+ */
+void pool_pending_messages_init(void)
+{
+	if (!session_context)
+		ereport(ERROR,
+				(errmsg("pool_pending_message_init: session context is not initialized")));
+
+	session_context->pending_messages = NIL;
+}
+
+/*
+ * Destroy pending message list
+ */
+void pool_pending_messages_destroy(void)
+{
+	ListCell   *cell;
+	POOL_PENDING_MESSAGE *msg;
+
+	if (!session_context)
+		ereport(ERROR,
+				(errmsg("pool_pending_message_destory: session context is not initialized")));
+
+	foreach(cell, session_context->pending_messages)
+	{
+		msg = (POOL_PENDING_MESSAGE *) lfirst(cell);
+		pfree(msg->contents);
+	}
+	list_free(session_context->pending_messages);
+}
+
+/*
+ * Create one message.
+ */
+POOL_PENDING_MESSAGE *pool_pending_messages_create(char kind, int len, char *contents)
+{
+	POOL_PENDING_MESSAGE* msg;
+	MemoryContext old_context;
+
+	if (!session_context)
+		ereport(ERROR,
+				(errmsg("pool_pending_message_create: session context is not initialized")));
+
+	old_context = MemoryContextSwitchTo(session_context->memory_context);
+	msg = palloc(sizeof(POOL_PENDING_MESSAGE));
+
+	switch (kind)
+	{
+		case 'P':
+		msg->type = POOL_PARSE;
+		break;
+
+		case 'B':
+		msg->type = POOL_BIND;
+		break;
+
+		case 'C':
+		msg->type = POOL_CLOSE;
+		break;
+
+		default:
+			ereport(ERROR,
+					(errmsg("pool_pending_message_create: unknow kind: %c", kind)));
+		break;
+	}
+
+	msg->contents = palloc(len);
+	memcpy(msg->contents, contents, len);
+	msg->contents_len = len;
+
+	MemoryContextSwitchTo(old_context);
+
+	return msg;
+}
+
+/*
+ * Add one message
+ */
+void pool_pending_message_add(POOL_PENDING_MESSAGE* message)
+{
+	POOL_PENDING_MESSAGE* msg;
+	MemoryContext old_context;
+
+	if (!session_context)
+		ereport(ERROR,
+				(errmsg("pool_pending_message_add: session context is not initialized")));
+
+	ereport(DEBUG1,
+			(errmsg("pool_pending_message_add: message type:%d message len:%d",
+					message->type, message->contents_len)));
+
+	old_context = MemoryContextSwitchTo(session_context->memory_context);
+	msg = copy_pending_message(message);
+	session_context->pending_messages = lappend(session_context->pending_messages, msg);
+	MemoryContextSwitchTo(old_context);
+}
+
+/*
+ * Try to find the first message specified by the message type in the message
+ * list. If found, a copy of the message is returned and the message is
+ * removed the message list. If not, returns NULL.
+ */
+POOL_PENDING_MESSAGE *pool_pending_message_remove(POOL_MESSAGE_TYPE type)
+{
+	ListCell   *cell;
+	ListCell   *prev;
+	ListCell   *next;
+	POOL_PENDING_MESSAGE *msg;
+	MemoryContext old_context;
+
+	if (!session_context)
+		ereport(ERROR,
+				(errmsg("pool_pending_message_remove: session context is not initialized")));
+
+	old_context = MemoryContextSwitchTo(session_context->memory_context);
+
+	prev = NULL;
+	msg = NULL;
+
+	for (cell = list_head(session_context->pending_messages); cell; cell = next)
+	{
+		POOL_PENDING_MESSAGE *m = (POOL_PENDING_MESSAGE *) lfirst(cell);
+
+		next = lnext(cell);
+
+		if (m->type == type)
+		{
+			msg = copy_pending_message(m);
+
+			session_context->pending_messages =
+				list_delete_cell(session_context->pending_messages, cell, prev);
+
+			break;
+		}
+		else
+			prev = cell;
+	}
+
+	MemoryContextSwitchTo(old_context);
+	return msg;
+}
+
+/*
+ * Get message specification (either statement ('S') or portal ('P')) from a
+ * close message.
+ */
+char pool_get_close_message_spec(POOL_PENDING_MESSAGE *msg)
+{
+	return *msg->contents;
+}
+
+/*
+ * Get statement or portal name from close message.
+ * The returned pointer is within "msg".
+ */
+char *pool_get_close_message_name(POOL_PENDING_MESSAGE *msg)
+{
+	return (msg->contents)+1;
+}
+
+/*
+ * Perform deep copy of POOL_PENDING_MESSAGE object in the current memory context.
+ */
+static POOL_PENDING_MESSAGE *copy_pending_message(POOL_PENDING_MESSAGE *message)
+{
+	POOL_PENDING_MESSAGE *msg;
+
+	msg = palloc(sizeof(POOL_PENDING_MESSAGE));
+	memcpy(msg, message, sizeof(POOL_PENDING_MESSAGE));
+	msg->contents = palloc(msg->contents_len);
+	memcpy(msg->contents, message->contents, msg->contents_len);
+
+	return msg;
+}
