@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2014	PgPool Global Development Group
+ * Copyright (c) 2003-2016	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -45,7 +45,6 @@
 
 #include "pool.h"
 #include "rewrite/pool_timestamp.h"
-#include "parallel_query/pool_rewrite_query.h"
 #include "rewrite/pool_lobj.h"
 #include "protocol/pool_proto_modules.h"
 #include "pool_config.h"
@@ -76,7 +75,6 @@ static int allow_close_transaction = 1;
 
 int is_select_pgcatalog = 0;
 int is_select_for_update = 0; /* 1 if SELECT INTO or SELECT FOR UPDATE */
-bool is_parallel_table = false;
 
 /*
  * last query string sent to simpleQuery()
@@ -89,7 +87,6 @@ static POOL_STATUS parse_before_bind(POOL_CONNECTION *frontend,
 									 POOL_CONNECTION_POOL *backend,
 									 POOL_SENT_MESSAGE *message);
 static int* find_victim_nodes(int *ntuples, int nmembers, int master_node, int *number_of_nodes);
-static int extract_ntuples(char *message);
 static POOL_STATUS close_standby_transactions(POOL_CONNECTION *frontend,
 											  POOL_CONNECTION_POOL *backend);
 /*
@@ -116,6 +113,8 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
 
+	bool error;
+
 	/* Get session context */
 	session_context = pool_get_session_context(false);
 
@@ -173,12 +172,12 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 	MemoryContext old_context = MemoryContextSwitchTo(query_context->memory_context);
 
 	/* parse SQL string */
-	parse_tree_list = raw_parser(contents);
+	parse_tree_list = raw_parser(contents, &error);
 
 	if (parse_tree_list == NIL)
 	{
 		/* is the query empty? */
-		if (*contents == '\0' || *contents == ';')
+		if (*contents == '\0' || *contents == ';' || error == false)
 		{
 			/*
 			 * JBoss sends empty queries for checking connections.
@@ -186,7 +185,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 			 * to affect load balance.
 			 * [Pgpool-general] Confused about JDBC and load balancing
 			 */
-			parse_tree_list = raw_parser(POOL_DUMMY_READ_QUERY);
+			parse_tree_list = raw_parser(POOL_DUMMY_READ_QUERY, &error);
 		}
 		else
 		{
@@ -209,7 +208,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 				ereport(LOG,
 						(errmsg("Unable to parse the query: \"%s\" from client %s(%s)", contents, remote_host, remote_port)));
 			}
-			parse_tree_list = raw_parser(POOL_DUMMY_WRITE_QUERY);
+			parse_tree_list = raw_parser(POOL_DUMMY_WRITE_QUERY, &error);
 			query_context->is_parse_error = true;
 		}
 	}
@@ -254,19 +253,6 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 			query_context->is_multi_statement = false;
 		}
 
-		if (PARALLEL_MODE)
-		{
-			bool parallel = true;
-
-			is_parallel_table = is_partition_table(backend,node);
-			status = pool_do_parallel_query(frontend, backend, node, &parallel, &contents, &len);
-			if (parallel)
-			{
-				pool_query_context_destroy(query_context);
-				return status;
-			}
-		}
-
 		/* check COPY FROM STDIN
 		 * if true, set copy_* variable
 		 */
@@ -397,37 +383,17 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 		 */
 		if (is_drop_database(node))
 		{
-			struct timeval stime;
+			int stime = 5;	/* XXX give arbitrary time to allow closing idle connections */
 
-			stime.tv_usec = 0;
-			stime.tv_sec = 5;	/* XXX give arbitrary time to allow
-									 * closing idle connections */
 			ereport(DEBUG1,
 					(errmsg("Query: sending SIGUSR1 signal to parent")));
-
-			ignore_sigusr1 = 1;	/* disable SIGUSR1 handler */
 			register_node_operation_request(CLOSE_IDLE_REQUEST, NULL, 0);
 
-			/*
-			 * We need to loop over here since we might get some signals while
-			 * sleeping
-			 */
-			for (;;)
+			/* we need to loop over here since we will get USR1 signal while sleeping */
+			while (stime > 0)
 			{
-				int sts;
-
-				errno = 0;
-				sts = select(0, NULL, NULL, NULL, &stime);
-				if (stime.tv_usec == 0 && stime.tv_sec == 0)
-					break;
-				if (sts != 0 && errno != EINTR)
-				{
-					elog(DEBUG1, "select(2) returns error: %s", strerror(errno));
-					break;
-				}
+				stime = sleep(stime);
 			}
-
-			ignore_sigusr1 = 0;
 		}
 
 		/*
@@ -439,7 +405,7 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 		 * - statement is INSERT
 		 * - either "INSERT LOCK" comment exists or insert_lock directive specified
 		 */
-		if (!RAW_MODE && !STREAM)
+		if (!RAW_MODE)
 		{
 			/*
 			 * If there's only one node to send the command, there's no
@@ -493,8 +459,6 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 										   query_context);
 			session_context->uncompleted_message =  msg;
 		}
-		else
-			session_context->uncompleted_message = NULL;
 	}
 
 
@@ -635,7 +599,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
         ereport(FATAL,
                 (return_code(2),
                  errmsg("unable to Execute"),
-                 errdetail("unable to get bind message")));
+                 errdetail("unable to get query context")));
 
 	if (!bind_msg->query_context->original_query)
         ereport(FATAL,
@@ -703,16 +667,6 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 				len += hexlen;
 			}
 
-			/*
-			 * If bind message is sent again to an existing prepared statement,
-			 * it is possible that query_w_hex remains. Before setting newly
-			 * allocated query_w_hex's pointer to the query context, free the
-			 * previously allocated memory.
-			 */
-			if (query_context->query_w_hex)
-			{
-				pfree(query_context->query_w_hex);
-			}
 			query_context->query_w_hex = search_query;
 
 			/*
@@ -762,7 +716,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	/* check if query is "COMMIT" or "ROLLBACK" */
 	commit = is_commit_or_rollback_query(node);
 
-	if (REPLICATION || PARALLEL_MODE)
+	if (!STREAM)
 	{
 		/*
 		 * Query is not commit/rollback
@@ -770,7 +724,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		if (!commit)
 		{
 			/* Send the query to master node */
-			pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID);
+			pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID, false);
 
 			/* Check specific errors */
 			specific_error = check_errors(backend, MASTER_NODE_ID);
@@ -789,23 +743,47 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			memset(msg + len, 0, sizeof(int));
 
 			/* send query to other nodes */
-			pool_extended_send_and_wait(query_context, "E", len, msg, -1, MASTER_NODE_ID);
+			pool_extended_send_and_wait(query_context, "E", len, msg, -1, MASTER_NODE_ID, false);
 		}
 		else
 		{
-			pool_extended_send_and_wait(query_context, "E", len, contents, -1, MASTER_NODE_ID);
+			pool_extended_send_and_wait(query_context, "E", len, contents, -1, MASTER_NODE_ID, false);
 		}
 		
 		/* send "COMMIT" or "ROLLBACK" to only master node if query is "COMMIT" or "ROLLBACK" */
 		if (commit)
 		{
-			pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID);
+			pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID, false);
 		}
 	}
-	else
+	else		/* streaming replication mode */
 	{
-		pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID);
-		pool_extended_send_and_wait(query_context, "E", len, contents, -1, MASTER_NODE_ID);
+		pool_extended_send_and_wait(query_context, "E", len, contents, 1, MASTER_NODE_ID, true);
+		pool_extended_send_and_wait(query_context, "E", len, contents, -1, MASTER_NODE_ID, true);
+
+		/*
+		 * Send flush message to backend to make sure that we get any response
+		 * from backend.
+		 */
+		pool_write(MASTER(session_context->backend), "H", 1);
+		len = htonl(sizeof(len));
+		pool_write_and_flush(MASTER(session_context->backend), &len, sizeof(len));
+		if (MASTER(session_context->backend)->db_node_id != session_context->load_balance_node_id)
+		{
+			POOL_CONNECTION *con;
+
+			con = session_context->backend->slots[session_context->load_balance_node_id]->con;
+			pool_write(con, "H", 1);
+			len = htonl(sizeof(len));
+			pool_write_and_flush(con, &len, sizeof(len));
+		}
+		/*
+		 * Remeber that we send flush or sync message to backend.
+		 */
+		pool_unset_pending_response();
+#ifdef NOT_USED
+		pool_unset_query_in_progress();
+#endif
 	}
 
 	return POOL_CONTINUE;
@@ -827,6 +805,8 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
 
+	bool error;
+
 	/* Get session context */
 	session_context = pool_get_session_context(false);
 
@@ -841,12 +821,12 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 
 	/* parse SQL string */
 	MemoryContext old_context = MemoryContextSwitchTo(query_context->memory_context);
-	parse_tree_list = raw_parser(stmt);
+	parse_tree_list = raw_parser(stmt, &error);
 
 	if (parse_tree_list == NIL)
 	{
 		/* is the query empty? */
-		if (*stmt == '\0' || *stmt == ';')
+		if (*stmt == '\0' || *stmt == ';' || error == false)
 		{
 			/*
 			 * JBoss sends empty queries for checking connections.
@@ -854,7 +834,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			 * to affect load balance.
 			 * [Pgpool-general] Confused about JDBC and load balancing
 			 */
-			parse_tree_list = raw_parser(POOL_DUMMY_READ_QUERY);
+			parse_tree_list = raw_parser(POOL_DUMMY_READ_QUERY, &error);
 		}
 		else
 		{
@@ -877,7 +857,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 				ereport(LOG,
 						(errmsg("Unable to parse the query: \"%s\" from client %s(%s)", stmt, remote_host, remote_port)));
 			}
-			parse_tree_list = raw_parser(POOL_DUMMY_WRITE_QUERY);
+			parse_tree_list = raw_parser(POOL_DUMMY_WRITE_QUERY, &error);
 			query_context->is_parse_error = true;
 		}
 	}
@@ -899,7 +879,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		/*
 		 * Start query context
 		 */
-		pool_start_query(query_context, stmt, strlen(stmt) + 1, node);
+		pool_start_query(query_context, pstrdup(stmt), strlen(stmt) + 1, node);
 
 		msg = pool_create_sent_message('P', len, contents, 0, name, query_context);
 
@@ -1074,7 +1054,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		}
 	}
 
-	if (REPLICATION || PARALLEL_MODE || MASTER_SLAVE)
+	if (REPLICATION || SLONY)
 	{
 		/*
 		 * We must synchronize because Parse message acquires table
@@ -1082,7 +1062,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		 */
 		ereport(DEBUG1,
 				(errmsg("Parse: waiting for master completing the query")));
-        pool_extended_send_and_wait(query_context, "P", len, contents, 1, MASTER_NODE_ID);
+        pool_extended_send_and_wait(query_context, "P", len, contents, 1, MASTER_NODE_ID, false);
 
 
 		/*
@@ -1123,12 +1103,22 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		}
 		else
 		{
-			pool_extended_send_and_wait(query_context, "P", len, contents, -1, MASTER_NODE_ID);
+			pool_extended_send_and_wait(query_context, "P", len, contents, -1, MASTER_NODE_ID, false);
 		}
 	}
+	else if (STREAM)
+	{
+		/* XXX fix me:even with streaming replication mode, couldn't we have a deadlock */
+		pool_set_query_in_progress();
+		pool_clear_sync_map();
+		pool_extended_send_and_wait(query_context, "P", len, contents, 1, MASTER_NODE_ID, true);
+		pool_extended_send_and_wait(query_context, "P", len, contents, -1, MASTER_NODE_ID, true);
+		pool_add_sent_message(session_context->uncompleted_message);
+		pool_unset_query_in_progress();
+	}
 	else
 	{
-		pool_extended_send_and_wait(query_context, "P", len, contents, 1, MASTER_NODE_ID);
+		pool_extended_send_and_wait(query_context, "P", len, contents, 1, MASTER_NODE_ID, false);
 	}
 
 	return POOL_CONTINUE;
@@ -1146,6 +1136,7 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
 	int insert_stmt_with_lock = 0;
+	bool nowait;
 
 	/* Get session context */
 	session_context = pool_get_session_context(false);
@@ -1197,6 +1188,10 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 
 	session_context->query_context = query_context;
 
+/*
+ * Fix me
+ */
+#ifdef NOT_USED
 	if (pool_config->load_balance_mode && pool_is_writing_transaction())
 	{
 		pool_where_to_send(query_context, query_context->original_query,
@@ -1205,6 +1200,7 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		if (parse_before_bind(frontend, backend, parse_msg) != POOL_CONTINUE)
 			return POOL_END;
 	}
+#endif
 
 	/*
 	 * Start a transaction if necessary in replication mode
@@ -1236,12 +1232,23 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			}
 		}
 	}
+
 	ereport(DEBUG1,
 		(errmsg("Bind: waiting for master completing the query")));
 
-	pool_extended_send_and_wait(query_context, "B", len, contents, 1, MASTER_NODE_ID);
+	pool_set_query_in_progress();
 
-	pool_extended_send_and_wait(query_context, "B", len, contents, -1, MASTER_NODE_ID);
+	nowait = (STREAM? true: false);
+
+	pool_clear_sync_map();
+	pool_extended_send_and_wait(query_context, "B", len, contents, 1, MASTER_NODE_ID, nowait);
+	pool_extended_send_and_wait(query_context, "B", len, contents, -1, MASTER_NODE_ID, nowait);
+
+	if (STREAM)
+	{
+		pool_unset_query_in_progress();
+		pool_add_sent_message(session_context->uncompleted_message);
+	}
 	
 	if(rewrite_msg)
 		pfree(rewrite_msg);
@@ -1255,6 +1262,8 @@ POOL_STATUS Describe(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
 
+	bool nowait;
+
 	/* Get session context */
 	session_context = pool_get_session_context(false);
 
@@ -1300,9 +1309,14 @@ POOL_STATUS Describe(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
     ereport(DEBUG1,
             (errmsg("Describe: waiting for master completing the query")));
 
-	pool_extended_send_and_wait(query_context, "D", len, contents, 1, MASTER_NODE_ID);
+	nowait = (REPLICATION? false: true);
 
-	pool_extended_send_and_wait(query_context, "D", len, contents, -1, MASTER_NODE_ID);
+	pool_set_query_in_progress();
+	pool_extended_send_and_wait(query_context, "D", len, contents, 1, MASTER_NODE_ID, nowait);
+	pool_extended_send_and_wait(query_context, "D", len, contents, -1, MASTER_NODE_ID, nowait);
+
+	if (!REPLICATION)
+		pool_unset_query_in_progress();
 
 	return POOL_CONTINUE;
 }
@@ -1335,9 +1349,10 @@ POOL_STATUS Close(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
                 (return_code(2),
                     errmsg("unable to execute close, invalid message")));
 	/* 
-	 * As per the postgresql, calling close on non existing portals is not
-	 * an error. So on the same footings we will ignore all such calls and
-	 * return the close complete message to clients with out going to backend
+	 * As per the postgresql, calling close on non existing portals or
+	 * statements is not an error. So on the same footings we will ignore all
+	 * such calls and return the close complete message to clients with out
+	 * going to backend
 	 */
 	if (!msg)
 	{
@@ -1365,9 +1380,26 @@ POOL_STATUS Close(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 
     ereport(DEBUG1,
             (errmsg("Close: waiting for master completing the query")));
-	pool_extended_send_and_wait(query_context, "C", len, contents, 1, MASTER_NODE_ID);
 
-	pool_extended_send_and_wait(query_context, "C", len, contents, -1, MASTER_NODE_ID);
+	pool_set_query_in_progress();
+
+	if (REPLICATION)
+	{
+		pool_extended_send_and_wait(query_context, "C", len, contents, 1, MASTER_NODE_ID, false);
+		pool_extended_send_and_wait(query_context, "C", len, contents, -1, MASTER_NODE_ID, false);
+	}
+	else
+	{
+		POOL_PENDING_MESSAGE *pmsg;
+
+		pool_clear_sync_map();
+		pool_extended_send_and_wait(query_context, "C", len, contents, 1, MASTER_NODE_ID, true);
+		pool_extended_send_and_wait(query_context, "C", len, contents, -1, MASTER_NODE_ID, true);
+
+		pmsg = pool_pending_messages_create('C', len, contents);
+		pool_pending_message_add(pmsg);
+		pool_unset_query_in_progress();
+	}
 
 	return POOL_CONTINUE;
 }
@@ -1413,10 +1445,19 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 	POOL_SESSION_CONTEXT *session_context;
 	Node *node = NULL;
 	char *query = NULL;
-	bool got_estate = false;
+	POOL_SYNC_MAP_STATE use_sync_map;
+
+	/*
+	 * It is possible that the "ignore until sync is received" flag was set if
+	 * we send sync to backend and the backend returns error. Let's reset the
+	 * flag unconditionally because we apparently have received a "ready for
+	 * query" message from backend.
+	 */
+	pool_unset_ignore_till_sync();
 
 	/* Get session context */
 	session_context = pool_get_session_context(false);
+	use_sync_map = pool_use_sync_map();
 
 	/*
 	 * If the numbers of update tuples are differ and
@@ -1475,7 +1516,7 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 				free_string(msg);
 
 				degenerate_backend_set(victim_nodes, number_of_nodes);
-				child_exit(1);
+				child_exit(POOL_EXIT_AND_RESTART);
 			}
 			else
 			{
@@ -1554,9 +1595,14 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 
 		for (i=0;i<NUM_BACKENDS;i++)
 		{
-			if (!VALID_BACKEND(i))
+			if (!VALID_BACKEND(i) || use_sync_map == POOL_SYNC_MAP_EMPTY)
 				continue;
 
+			if (use_sync_map == POOL_SYNC_MAP_IS_VALID && !pool_is_set_sync_map(i))
+			{
+				continue;
+			}
+
 			if (pool_read(CONNECTION(backend, i), &kind, sizeof(kind)))
 				return POOL_END;
 
@@ -1572,13 +1618,6 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 			{
 				state = kind;
 			}
-			/*
-			 * However, if the state is 'E', then frontend should had been
-			 * already reported an ERROR. So, to match with that, let be the
-			 * state to be returned to frontend.
-			 */
-			if (kind == 'E')
-				got_estate = true;
 		}
 	}
 
@@ -1590,8 +1629,6 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 		{
 			len = htonl(len);
 			pool_write(frontend, &len, sizeof(len));
-			if (got_estate)
-				state = 'E';
 			pool_write(frontend, &state, 1);
 		}
 		pool_flush(frontend);
@@ -1736,7 +1773,6 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 		else
 		{
 			if ((pool_is_doing_extended_query_message() &&
-				 session_context->query_context &&
 				 session_context->query_context->query_state[MASTER_NODE_ID] != POOL_UNPARSED &&
 			     session_context->uncompleted_message) ||
 			    (!pool_is_doing_extended_query_message() && session_context->uncompleted_message))
@@ -1843,6 +1879,8 @@ POOL_STATUS CloseComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backe
 {
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_STATUS status;
+	char kind = ' ';
+	char *name = "";
 
 	/* Get session context */
 	session_context = pool_get_session_context(false);
@@ -1851,254 +1889,44 @@ POOL_STATUS CloseComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backe
 	status = SimpleForwardToFrontend('3', frontend, backend);
 
 	/* Remove the target message */
-	if (session_context->uncompleted_message)
+	if (STREAM)
 	{
-		pool_remove_sent_message(session_context->uncompleted_message->kind,
-								 session_context->uncompleted_message->name);
-		session_context->uncompleted_message = NULL;
+		POOL_PENDING_MESSAGE *pmsg;
+
+		pmsg = pool_pending_message_remove(POOL_CLOSE);
+
+		if (pmsg)
+		{
+			kind = pool_get_close_message_spec(pmsg);
+			name = pool_get_close_message_name(pmsg);
+		}
 	}
 	else
 	{
-		ereport(ERROR,
-				(errmsg("processing CloseComplete, uncompleted message not found")));
+		if (session_context->uncompleted_message)
+		{
+			kind = session_context->uncompleted_message->kind;
+			name = session_context->uncompleted_message->name;
+			session_context->uncompleted_message = NULL;
+		}
+		else
+		{
+			ereport(ERROR,
+					(errmsg("processing CloseComplete, uncompleted message not found")));
+		}
+	}
+	
+	if (kind == ' ')
+	{
+		pool_remove_sent_message(kind, name);
+		ereport(DEBUG1,
+				(errmsg("CloseComplete: remove uncompleted message. kind:%c, name:%s",
+						kind, name)));
 	}
 
 	return status;
 }
 
-POOL_STATUS CommandComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend)
-{
-	int i;
-	int len, len1, sendlen;
-	int rows;
-	char *p, *p1, *p2;
-	bool update_or_delete = false;
-	POOL_SESSION_CONTEXT *session_context;
-
-	/* Get session context */
-	session_context = pool_get_session_context(false);
-
-	if (session_context->query_context != NULL)
-	{
-		Node *node = session_context->query_context->parse_tree;
-
-		if (IsA(node, PrepareStmt))
-		{
-			if (session_context->uncompleted_message)
-			{
-				pool_add_sent_message(session_context->uncompleted_message);
-				session_context->uncompleted_message = NULL;
-			}
-		}
-		else if (IsA(node, DeallocateStmt))
-		{
-			char *name;
-			
-			name = ((DeallocateStmt *)node)->name;
-			if (name == NULL)
-			{
-				pool_remove_sent_messages('Q');
-				pool_remove_sent_messages('P');
-			}
-			else
-			{
-				pool_remove_sent_message('Q', name);
-				pool_remove_sent_message('P', name);
-			}
-		}
-		else if (IsA(node, DiscardStmt))
-		{
-			DiscardStmt *stmt = (DiscardStmt *)node;
-
-			if (stmt->target == DISCARD_PLANS)
-			{
-				pool_remove_sent_messages('Q');
-				pool_remove_sent_messages('P');
-			}
-			else if (stmt->target == DISCARD_ALL)
-			{
-				pool_clear_sent_message_list();
-			}
-		}
-		/*
-		 * JDBC driver sends "BEGIN" query internally if
-		 * setAutoCommit(false).  But it does not send Sync message
-		 * after "BEGIN" query.  In extended query protocol,
-		 * PostgreSQL returns ReadyForQuery when a client sends Sync
-		 * message.  Problem is, pgpool can't know the transaction
-		 * state without receiving ReadyForQuery. So we remember that
-		 * we need to send Sync message internally afterward, whenever
-		 * we receive BEGIN in extended protocol.
-		 */
-		else if (IsA(node, TransactionStmt))
-		{
-			TransactionStmt *stmt = (TransactionStmt *) node;
-
-			if (stmt->kind == TRANS_STMT_BEGIN || stmt->kind == TRANS_STMT_START)
-			{
-				int i;
-
-				for (i = 0; i < NUM_BACKENDS; i++)
-				{
-					if (!VALID_BACKEND(i))
-						continue;
-
-					TSTATE(backend, i) = 'T';
-				}
-				pool_unset_writing_transaction();
-				pool_unset_failed_transaction();
-				pool_unset_transaction_isolation();
-			}
-		}
-
-	}
-
-	pool_read(MASTER(backend), &len, sizeof(len));
-
-	len = ntohl(len);
-	len -= 4;
-	len1 = len;
-
-	p = pool_read2(MASTER(backend), len);
-	if (p == NULL)
-		return POOL_END;
-	p1 = palloc(len);
-	memcpy(p1, p, len);
-
-	rows = extract_ntuples(p);
-
-	/*
-	 * Save number of affected tuples of master node.
-	 */
-	session_context->ntuples[MASTER_NODE_ID] = rows;
-
-	if (strstr(p, "UPDATE") || strstr(p, "DELETE"))
-		update_or_delete = true;
-
-	for (i=0;i<NUM_BACKENDS;i++)
-	{
-		if (!IS_MASTER_NODE_ID(i))
-		{
-			if (!VALID_BACKEND(i))
-			{
-				session_context->ntuples[i] = -1;
-				continue;
-			}
-
-			pool_read(CONNECTION(backend, i), &len, sizeof(len));
-
-			len = ntohl(len);
-			len -= 4;
-
-			p = pool_read2(CONNECTION(backend, i), len);
-			if (p == NULL)
-				return POOL_END;
-
-			if (len != len1)
-			{
-				ereport(DEBUG1,
-					(errmsg("processing command complete"),
-						errdetail("length does not match between backends master(%d) %d th backend(%d)",
-							   len, i, len1)));
-			}
-
-			int n = extract_ntuples(p);
-
-			/*
-			 * Save number of affected tuples.
-			 */
-			session_context->ntuples[i] = n;
-
-			/*
-			 * if we are in the parallel mode, we have to sum up the number
-			 * of affected rows
-			 */
-			if (PARALLEL_MODE && is_parallel_table && update_or_delete)
-			{
-				rows += n;
-			}
-			else
-			{
-				if (rows != n)
-				{
-					/*
-					 * Remember that we have different number of UPDATE/DELETE
-					 * affected tuples in backends.
-					 */
-					session_context->mismatch_ntuples = true;
-				}
-			}
-		}
-	}
-
-	if (session_context->mismatch_ntuples)
-	{
-		char msgbuf[128];
-
-		String *msg = init_string("pgpool detected difference of the number of inserted, updated or deleted tuples. Possible last query was: \"");
-		string_append_char(msg, query_string_buffer);
-		string_append_char(msg, "\"");
-		pool_send_error_message(frontend, MAJOR(backend),
-								"XX001", msg->data, "",
-								"check data consistency between master and other db node",  __FILE__, __LINE__);
-		ereport(LOG,
-			(errmsg("%s", msg->data)));
-		free_string(msg);
-
-		msg = init_string("CommandComplete: Number of affected tuples are:");
-
-		for (i=0;i<NUM_BACKENDS;i++)
-		{
-			snprintf(msgbuf, sizeof(msgbuf), " %d", session_context->ntuples[i]);
-			string_append_char(msg, msgbuf);
-		}
-		ereport(LOG,
-			(errmsg("processing command complete"),
-				 errdetail("%s", msg->data)));
-
-		free_string(msg);
-	}
-	else
-	{
-		if (PARALLEL_MODE && is_parallel_table && update_or_delete)
-		{
-			char tmp[32];
-
-			strncpy(tmp, p1, 7);
-			sprintf(tmp+7, "%d", rows);
-
-			p2 = pstrdup(tmp);
-
-			pfree(p1);
-			p1 = p2;
-			len1 = strlen(p2) + 1;
-		}
-
-		pool_write(frontend, "C", 1);
-		sendlen = htonl(len1+4);
-		pool_write(frontend, &sendlen, sizeof(sendlen));
-		pool_write_and_flush(frontend, p1, len1);
-	}
-
-	/* Save the received result to buffer for each kind */
-	if (pool_config->memory_cache_enabled)
-	{
-		if (pool_is_cache_safe() && !pool_is_cache_exceeded())
-		{
-			memqcache_register('C', frontend, p1, len1);
-		}
-	}
-
-	pfree(p1);
-
-	if (pool_is_doing_extended_query_message())
-	{
-		pool_set_query_state(session_context->query_context, POOL_EXECUTE_COMPLETE);
-	}
-
-	return POOL_CONTINUE;
-}
-
 POOL_STATUS ParameterDescription(POOL_CONNECTION *frontend,
 								 POOL_CONNECTION_POOL *backend)
 {
@@ -2425,6 +2253,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 		char *query;
 		Node *node;
 		List *parse_tree_list;
+		bool error;
 
 		case 'X':	/* Terminate */
 			if(contents)
@@ -2450,29 +2279,39 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 		case 'P':	/* Parse */
 			allow_close_transaction = 0;
 			pool_set_doing_extended_query_message();
+#ifdef NOT_USED
 			if (!pool_is_query_in_progress() && !pool_is_ignore_till_sync())
 				pool_set_query_in_progress();
+#endif
 			status = Parse(frontend, backend, len, contents);
+			pool_set_pending_response();
 			break;
 
 		case 'B':	/* Bind */
 			pool_set_doing_extended_query_message();
+#ifdef NOT_USED
 			if (!pool_is_query_in_progress() && !pool_is_ignore_till_sync())
 				pool_set_query_in_progress();
+#endif
 			status = Bind(frontend, backend, len, contents);
+			pool_set_pending_response();
 			break;
 
 		case 'C':	/* Close */
 			if (!pool_is_query_in_progress() && !pool_is_ignore_till_sync())
 				pool_set_query_in_progress();
 			status = Close(frontend, backend, len, contents);
+			pool_set_pending_response();
 			break;
 
 		case 'D':	/* Describe */
 			pool_set_doing_extended_query_message();
+#ifdef NOT_USED
 			if (!pool_is_query_in_progress() && !pool_is_ignore_till_sync())
 				pool_set_query_in_progress();
+#endif
 			status = Describe(frontend, backend, len, contents);
+			pool_set_pending_response();
 			break;
 
 		case 'S':  /* Sync */
@@ -2482,6 +2321,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 			if (!pool_is_query_in_progress())
 				pool_set_query_in_progress();
 			status = SimpleForwardToBackend(fkind, frontend, backend, len, contents);
+			pool_unset_pending_response();
 			break;
 
 		case 'F':	/* FunctionCall */
@@ -2492,7 +2332,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 			query = "INSERT INTO foo VALUES(1)";
 			MemoryContext old_context = MemoryContextSwitchTo(query_context->memory_context);
 
-			parse_tree_list = raw_parser(query);
+			parse_tree_list = raw_parser(query, &error);
 			node = (Node *) lfirst(list_head(parse_tree_list));
 			pool_start_query(query_context, query, strlen(query) + 1, node);
 
@@ -2550,7 +2390,16 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 	{
 		if (pool_is_query_in_progress())
 			pool_unset_query_in_progress();
-		return POOL_CONTINUE;
+
+		/*
+		 * Check if If we have pending data in backend connection cache. If we
+		 * do, it is likely that a sync message has been sent to backend and the
+		 * backend replied back to us. So we need to process it.
+		 */
+		if (is_backend_cache_empty(backend))
+		{
+			return POOL_CONTINUE;
+		}
 	}
 
 	if (pool_is_skip_reading_from_backends())
@@ -2597,19 +2446,22 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 			case '1':	/* ParseComplete */
 				status = ParseComplete(frontend, backend);
 				pool_set_command_success();
-				pool_unset_query_in_progress();
+				if (REPLICATION||RAW_MODE)
+					pool_unset_query_in_progress();
 				break;
 
 			case '2':	/* BindComplete */
 				status = BindComplete(frontend, backend);
 				pool_set_command_success();
-				pool_unset_query_in_progress();
+				if (REPLICATION||RAW_MODE)
+					pool_unset_query_in_progress();
 				break;
 
 			case '3':	/* CloseComplete */
 				status = CloseComplete(frontend, backend);
 				pool_set_command_success();
-				pool_unset_query_in_progress();
+				if (REPLICATION||RAW_MODE)
+					pool_unset_query_in_progress();
 				break;
 
 			case 'E':	/* ErrorResponse */
@@ -2628,7 +2480,7 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 			case 'C':	/* CommandComplete */				
 				status = CommandComplete(frontend, backend);
 				pool_set_command_success();
-				if (pool_is_doing_extended_query_message())
+				if ((REPLICATION || RAW_MODE) && pool_is_doing_extended_query_message())
 					pool_unset_query_in_progress();
 				break;
 
@@ -2638,25 +2490,25 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 
 			case 'I':	/* EmptyQueryResponse */
 				status = SimpleForwardToFrontend(kind, frontend, backend);
-				if (pool_is_doing_extended_query_message())
+				if ((REPLICATION || RAW_MODE) && pool_is_doing_extended_query_message())
 					pool_unset_query_in_progress();
 				break;
 
 			case 'T':	/* RowDescription */
 				status = SimpleForwardToFrontend(kind, frontend, backend);
-				if (pool_is_doing_extended_query_message())
+				if ((REPLICATION || RAW_MODE) && pool_is_doing_extended_query_message())
 					pool_unset_query_in_progress();
 				break;
 
 			case 'n':	/* NoData */
 				status = SimpleForwardToFrontend(kind, frontend, backend);
-				if (pool_is_doing_extended_query_message())
+				if ((REPLICATION || RAW_MODE) && pool_is_doing_extended_query_message())
 					pool_unset_query_in_progress();
 				break;
 
 			case 's':	/* PortalSuspended */
 				status = SimpleForwardToFrontend(kind, frontend, backend);
-				if (pool_is_doing_extended_query_message())
+				if ((REPLICATION || RAW_MODE) && pool_is_doing_extended_query_message())
 					pool_unset_query_in_progress();
 				break;
 
@@ -2791,20 +2643,12 @@ POOL_STATUS CopyDataRows(POOL_CONNECTION *frontend,
 	char *string = NULL;
 	int len;
 	int i;
-	DistDefInfo *info = NULL;
 
 #ifdef DEBUG
 	int j = 0;
 	char buf[1024];
 #endif
 
-	if (copyin && pool_config->parallel_mode == TRUE)
-	{
-		info = pool_get_dist_def_info(MASTER_CONNECTION(backend)->sp->database,
-									  copy_schema,
-									  copy_table);
-	}
-
 	for (;;)
 	{
 		if (copyin)
@@ -2812,86 +2656,20 @@ POOL_STATUS CopyDataRows(POOL_CONNECTION *frontend,
 			if (MAJOR(backend) == PROTO_MAJOR_V3)
 			{
 				char kind;
-				int sendlen;
-				char *p, *p1;
+				char *contents = NULL;
 
 				pool_read(frontend, &kind, 1);
 
-				if (info && kind == 'd')
-				{
-					int id;
-					pool_read(frontend, &sendlen, sizeof(sendlen));
+				ereport(DEBUG1,
+					(errmsg("copy data rows"),
+						 errdetail("read kind from frontend %c(%02x)", kind, kind)));
 
-					len = ntohl(sendlen) - 4;
+				pool_read(frontend, &len, sizeof(len));
+				len = ntohl(len) - 4;
+				if (len > 0)
+					contents = pool_read2(frontend, len);
 
-					if (len <= 0)
-						return POOL_CONTINUE;
-
-					p = pool_read2(frontend, len);
-					if (p == NULL)
-                        ereport(ERROR,
-                            (errmsg("unable to copy data rows"),
-                                 errdetail("read on backend node failed")));
-
-					/* copy end ? */
-					if (len == 3 && memcmp(p, "\\.\n", 3) == 0)
-					{
-						for (i=0;i<NUM_BACKENDS;i++)
-						{
-							if (VALID_BACKEND(i))
-							{
-								pool_write(CONNECTION(backend, i), &kind, 1);
-								pool_write(CONNECTION(backend, i), &sendlen, sizeof(sendlen));
-								pool_write(CONNECTION(backend, i), p, len);
-							}
-						}
-					}
-					else
-					{
-						p1 = parse_copy_data(p, len, copy_delimiter, info->dist_key_col_id);
-
-						if (!p1)
-	                         ereport(ERROR,
-                                (errmsg("unable to copy data rows"),
-                                     errdetail("unable to parse data")));
-   
-						else if (strcmp(p1, copy_null) == 0)
-                            ereport(ERROR,
-                                    (errmsg("unable to copy data rows"),
-                                     errdetail("key parameter is NULL")));
-
-						id = pool_get_id(info, p1);
-
-						ereport(DEBUG1,
-							(errmsg("copy data rows"),
-								 errdetail("copying id: %d", id)));
-						pfree(p1);
-						if (id < 0 || !VALID_BACKEND(id))
-                            ereport(FATAL,
-								(return_code(1),
-                                    errmsg("unable to copy data rows"),
-                                     errdetail("pool_get_id returns invalid id: %d", id)));
-                                    
-
-                        pool_write(CONNECTION(backend, id), &kind, 1);
-                        pool_write(CONNECTION(backend, id), &sendlen, sizeof(sendlen));
-                        pool_write_and_flush(CONNECTION(backend, id), p, len);
-					}
-				}
-				else
-				{
-					char *contents = NULL;
-					ereport(DEBUG1,
-						(errmsg("copy data rows"),
-							 errdetail("read kind from frontend %c(%02x)", kind, kind)));
-
-					pool_read(frontend, &len, sizeof(len));
-					len = ntohl(len) - 4;
-					if (len > 0)
-						contents = pool_read2(frontend, len);
-
-					SimpleForwardToBackend(kind, frontend, backend, len, contents);
-				}
+				SimpleForwardToBackend(kind, frontend, backend, len, contents);
 
 				/* CopyData? */
 				if (kind == 'd')
@@ -3190,12 +2968,11 @@ void per_node_error_log(POOL_CONNECTION_POOL *backend, int node_id, char *query,
 	POOL_CONNECTION_POOL_SLOT *slot = backend->slots[node_id];
 	char *message;
 
-	if (pool_extract_error_message(true, CONNECTION(backend, node_id), MAJOR(backend), true, &message) == 1)
+	if (pool_extract_error_message(true, CONNECTION(backend, node_id), MAJOR(backend), true, &message) > 0)
 	{
 		ereport(LOG,
 			(errmsg("%s: DB node id: %d backend pid: %d statement: \"%s\" message: \"%s\"",
 				 prefix, node_id, ntohl(slot->pid), query, message)));
-		pfree(message);
 	}
 }
 
@@ -3222,7 +2999,7 @@ static POOL_STATUS parse_before_bind(POOL_CONNECTION *frontend,
 				(errmsg("parse before bind"),
 					 errdetail("waiting for backend %d completing parse", i)));
 
-			pool_extended_send_and_wait(qc, "P", len, contents, 1, i);
+			pool_extended_send_and_wait(qc, "P", len, contents, 1, i, false);
 		}
 		else
 		{
@@ -3241,8 +3018,6 @@ static POOL_STATUS parse_before_bind(POOL_CONNECTION *frontend,
 	
 	if (parse_was_sent)
 	{
-		pool_set_query_in_progress();
-
 		while (kind != '1')
 		{
 			PG_TRY();
@@ -3376,23 +3151,3 @@ static int* find_victim_nodes(int *ntuples, int nmembers, int master_node, int *
 
 	return victim_nodes;
 }
-
-/*
- * Extract the number of tuples from CommandComplete message
- */
-static int extract_ntuples(char *message)
-{
-	char *rows;
-
-	if ((rows = strstr(message, "UPDATE")) || (rows = strstr(message, "DELETE")))
-		rows +=7;
-	else if ((rows = strstr(message, "INSERT")))
-	{
-		rows += 7;
-		while (*rows && *rows != ' ') rows++;
-	}
-	else
-		return 0;
-
-	return atoi(rows);
-}
